"""
This file is part of pysofar: A client for interfacing with Sofar Oceans Spotter API

Contents: 2D Spectrum Class

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
"""
import numpy
import numpy.ma
from .wavespectrum import WaveSpectrum, WaveSpectrumInput
from typing import List, Union
from roguewave.tools import datetime_to_iso_time_string
from numpy.ma import MaskedArray
from numba import njit
from functools import cached_property

class WaveSpectrum2DInput(WaveSpectrumInput):
    directions: Union[List[float], numpy.ndarray]

class WaveSpectrum2D(WaveSpectrum):
    def __init__(self,
                 wave_spectrum2D_input:WaveSpectrum2DInput
                 ):
        super().__init__(wave_spectrum2D_input)
        self.direction = numpy.array(wave_spectrum2D_input['directions'],
                                     dtype='float64')
        self._directional_difference = self._delta()

        self._frequency_peak_indices = None
        self._direction_peak_indices = None

    @cached_property
    def e(self) -> numpy.array:
        return _directional_moment( self.variance_density.filled(0),
                                     self.radian_direction,
                                     self._directional_difference,
                                     'zero',
                                     0,
                                     False
                                    )

    @cached_property
    def a1(self) -> numpy.array:
        return _directional_moment( self.variance_density.filled(0),
                                    self.radian_direction,
                                    self._directional_difference,
                                    'a',
                                    1,
                                    True
        )

    @cached_property
    def b1(self) -> numpy.array:
        return _directional_moment(self.variance_density.filled(0),
                                   self.radian_direction,
                                   self._directional_difference,
                                   'b',
                                   1,
                                   True
        )


    @cached_property
    def a2(self) -> numpy.array:
        return _directional_moment( self.variance_density.filled(0),
                                    self.radian_direction,
                                    self._directional_difference,
                                    'a',
                                    2,
                                    True
        )

    @cached_property
    def b2(self) -> numpy.array:
        return _directional_moment(self.variance_density.filled(0),
                                   self.radian_direction,
                                   self._directional_difference,
                                   'b',
                                   2,
                                   True
                                   )


    @WaveSpectrum.variance_density.setter
    def variance_density(self,val:numpy.ndarray):
        val = val.copy()
        mask = val < 0
        self._variance_density = MaskedArray(val,mask=mask,dtype='float64')
        self._update()

    def _update(self):
        super(WaveSpectrum2D, self)._update()

        if hasattr(self,'e'): delattr(self,'e')
        if hasattr(self,'a1'):    delattr(self,'a1')
        if hasattr(self,'b1'):    delattr(self,'b1')
        if hasattr(self,'a2'):    delattr(self,'a2')
        if hasattr(self,'b2'):    delattr(self,'b2')

        #self._e = None

    def _delta(self):
        angles = self.direction
        forward_diff = (numpy.diff(angles, append=angles[0]) + 180) % 360 - 180
        backward_diff = (numpy.diff(angles,
                                    prepend=angles[-1]) + 180) % 360 - 180
        return (forward_diff + backward_diff) / 2


    def frequency_moment(self, power: int, fmin=0, fmax=numpy.inf) -> float:
        range = self._range(fmin, fmax)
        return numpy.trapz(self.e[range] * self.frequency[range] ** power,
                           self.frequency[range])

    def _create_wave_spectrum_input(self)->WaveSpectrum2DInput:
        data = self.variance_density.filled(-1)
        return WaveSpectrum2DInput(
            frequency=self.frequency,
            directions=self.direction,
            varianceDensity=data,
            timestamp=datetime_to_iso_time_string(self.timestamp),
            latitude=self.latitude,
            longitude=self.longitude,
        )

    def copy(self)->"WaveSpectrum2D":
        input = WaveSpectrum2DInput(
            frequency=self.frequency.copy(),
            directions=self.direction.copy(),
            varianceDensity=self.variance_density.copy(),
            timestamp=self.timestamp,
            latitude=self.latitude,
            longitude=self.longitude
        )
        return WaveSpectrum2D(input)

    def __add__(self, other:"WaveSpectrum2D")->"WaveSpectrum2D":
        spectrum = self.copy()

        self_dens = numpy.ma.MaskedArray(spectrum.variance_density)
        other_dens = numpy.ma.MaskedArray(other.variance_density)

        mask = self_dens.mask & other_dens.mask

        spectrum.variance_density = numpy.ma.MaskedArray(self_dens.filled(0) +
            other_dens.filled(0), mask)
        return spectrum

    def __sub__(self, other: "WaveSpectrum2D") -> "WaveSpectrum2D":
        spectrum = self.copy()
        self_dens = numpy.ma.MaskedArray(spectrum.variance_density)
        other_dens = numpy.ma.MaskedArray(other.variance_density)

        mask = self_dens.mask & other_dens.mask

        spectrum.variance_density = numpy.ma.MaskedArray(self_dens.filled(0) -
            other_dens.filled(0), mask)
        return spectrum

    def __neg__(self, other: "WaveSpectrum2D") -> "WaveSpectrum2D":
        spectrum = self.copy()
        spectrum.variance_density = -spectrum.variance_density
        return spectrum

    def extract(self,mask:numpy.ndarray)->"WaveSpectrum2D":
        spectrum = self.copy()
        density = spectrum.variance_density
        density[~mask] = 0
        density = numpy.ma.masked_array( density, mask=~mask)
        spectrum.variance_density = density
        return spectrum

    def peak_wavenumber_east(self, depth=numpy.inf):
        wave_number    = self.peak_wavenumber()
        wave_direction = self.peak_direction() * numpy.pi/180
        return wave_number * numpy.cos(wave_direction)

    def peak_wavenumber_north(self, depth=numpy.inf):
        wave_number    = self.peak_wavenumber()
        wave_direction = self.peak_direction() * numpy.pi/180
        return wave_number * numpy.cos(wave_direction)


def empty_spectrum2D_like(spectrum:WaveSpectrum2D)->WaveSpectrum2D:
    input = WaveSpectrum2DInput(
        frequency=spectrum.frequency.copy(),
        varianceDensity=numpy.zeros_like(spectrum.variance_density),
        timestamp=spectrum.timestamp,
        latitude=spectrum.latitude,
        longitude=spectrum.longitude,
        directions=spectrum.direction.copy()
    )
    return WaveSpectrum2D(input)


@njit(cache=True)
def _directional_moment(
        density:numpy.ndarray,
        radian_direction:numpy.ndarray,
        directional_difference:numpy.ndarray,
        kind:str='zero',
        order:int=0,
        normalized:bool=True) -> numpy.array:

    """
    Calculate the directional moments of a directional Spectrum

    :param density: 2D variance density
    :param radian_direction: radian angles
    :param directional_difference: centered difference between directions
    :param kind: which Fourier moment to calculate: a=cosine, b=sine
    :param order: which order of Fourier moment to calculate
    :param normalized: whether to return normalized or non-normalized moments
    :return:
    """

    if kind == 'a':
        harmonic = numpy.cos(radian_direction * order) * directional_difference
    elif kind == 'b':
        harmonic = numpy.sin(radian_direction * order) * directional_difference
    elif kind == 'zero':
        harmonic = directional_difference
    else:
        raise Exception('Unknown moment')

    values = numpy.sum(density * harmonic, axis=-1)

    if normalized:
        scale = numpy.sum(density * directional_difference, axis=-1)
        scale[scale == 0] = numpy.float64(1.0)
    else:
        scale = numpy.ones( values.shape[0] )
    return values / scale



import logging
import sys
logging.getLogger(__name__).addHandler(logging.NullHandler())
logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)
logger.addHandler(logging.NullHandler())

# External API tools

from .modeldata import sofarspectralapi
from roguewave.modeldata.sofarspectralapi import load_sofar_spectral_file

# Wave spectrum

## object definition
from .wavespectra.spectrum1D import WaveSpectrum1D, WaveSpectrum1DInput
from .wavespectra.spectrum2D import WaveSpectrum2D, WaveSpectrum2DInput
from .wavespectra import spectrum1D, spectrum2D

## estimators
from .wavespectra.estimators import convert_to_1d_spectrum,convert_to_2d_spectrum

## save/load
from .wavespectra.io import load_spectrum, save_spectrum # Depricated
from .io.io import save,load

## Partitioning
from .wavespectra.partitioning.wavefields import (
    get_bulk_partitions_from_spectral_partitions,
    get_spectral_partitions_from_2dspectra)
from .wavespectra.partitioning.observations import \
    get_bulk_partitions_from_observations, \
    get_spectral_partitions_from_observations

def set_log_to_console(level=logging.INFO):
    logger.addHandler(logging.StreamHandler(stream=sys.stdout))
    logger.setLevel(level)

def set_log_to_file(filename, level=logging.INFO):
    logger.addHandler(logging.FileHandler(filename))
    logger.setLevel(level)


from dataclasses import dataclass, field
from datetime import datetime, timezone
from typing import List

import numpy
from pandas import DataFrame

from roguewave.tools import to_datetime


@dataclass
class MetoceanData():
    latitude: float = numpy.nan
    longitude: float = numpy.nan
    timestamp: datetime = datetime(2022, 1, 1, tzinfo=timezone.utc)

    _timestamp: datetime = field(init=False, repr=False)

    @property
    def timestamp(self) -> datetime:
        return self._timestamp

    @timestamp.setter
    def timestamp(self, time):
        if isinstance(time, property):
            # Honestly not sure what is going on- but if we initialze
            # timestamp with the default value of None a "property" object
            # gets passed instead of just "None". This catches that and
            # makes sure it all works. Bit of an edge case, should not
            # happen if initialized with value. This seems to be due to
            # the wave the dataclass and getters and setters interact and
            # the magic involved to make it work.
            time = time.fdel
        self._timestamp = to_datetime(time)

    def as_dict(self):
        return {
            "latitude": self.latitude,
            "longitude": self.longitude,
            "timestamp": self.timestamp
            }

@dataclass
class WaveBulkData(MetoceanData):

    significant_waveheight: float = numpy.nan
    peak_period: float = numpy.nan
    mean_period: float = numpy.nan
    peak_direction: float = numpy.nan
    peak_directional_spread: float = numpy.nan
    mean_direction: float = numpy.nan
    mean_directional_spread: float = numpy.nan
    peak_frequency: float = numpy.nan

    def as_dict(self):
        return {
            "latitude": self.latitude,
            "longitude": self.longitude,
            "timestamp": self.timestamp,
            "significant_waveheight": self.significant_waveheight,
            "peak_period": self.peak_period,
            "mean_period": self.mean_period,
            "peak_direction": self.peak_direction,
            "peak_directional_spread": self.peak_directional_spread,
            "mean_direction": self.mean_direction,
            "mean_directional_spread": self.mean_directional_spread,
            "peak_frequency": self.peak_frequency
        }


@dataclass
class SSTData(MetoceanData):
    degrees: float = numpy.nan

    def as_dict(self):
        return {
            "latitude": self.latitude,
            "longitude": self.longitude,
            "timestamp": self.timestamp,
            "degrees": self.degrees,
        }

@dataclass
class WindData(MetoceanData):
    speed: float=numpy.nan
    direction: float = numpy.nan
    seasurfaceId: float = numpy.nan

    def as_dict(self):
        return {
            "latitude": self.latitude,
            "longitude": self.longitude,
            "timestamp": self.timestamp,
            "speed": self.speed,
            "direction": self.direction,
            "seasurfaceId": self.seasurfaceId
        }

@dataclass
class BarometricPressure(MetoceanData):
    units: str = ''
    value:float = numpy.nan
    unit_type: str = ''
    data_type_name: str = ''

    def as_dict(self):
        return {
            "latitude": self.latitude,
            "longitude": self.longitude,
            "timestamp": self.timestamp,
            "units": self.units,
            "unit_type": self.unit_type,
            "data_type_name": self.data_type_name,
            "value": self.value
        }


def as_dataframe(bulk_wave_properties:List[MetoceanData]):
    data = {}
    for bulk_propererties in bulk_wave_properties:
        dictionary = bulk_propererties.as_dict()
        for key in dictionary:
            if key not in data:
                data[key] = []
            data[key].append( dictionary[key])

    for key in data:
        data[key] = numpy.array(data[key])

    df= DataFrame.from_dict(data)
    df.set_index('timestamp', inplace=True)
    return df


"""
Contents: IO

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Routines that can be used to save and load Spectral data.

Functions:

- `load_spectrum`, load spectral data
- `save_spectrum`, save spectral data
- `_deserialize`, deserialize data
- `_serialize`, serialize data


How To Use This Module
======================
(See the individual functions for details.)

1. No need for direct import, use roguewave ``import roguewave`` or ``from roguewave import ...``.
2. Save data: save_spectrum(data, filename)
3. load data: data = load_spectrum(filename)

"""
from roguewave.wavespectra.spectrum1D import WaveSpectrum1D
from roguewave.wavespectra.spectrum2D import WaveSpectrum2D
from roguewave.metoceandata import WaveBulkData, WindData, SSTData,BarometricPressure
from roguewave import spectrum1D, spectrum2D
from pandas import DataFrame, read_json
from typing import Union, Dict, List
from datetime import datetime
import gzip
import json
import os
import numpy
import base64


_UNION = Union[
    WaveSpectrum1D,
    WaveSpectrum2D,
    List[WaveSpectrum1D],
    List[WaveSpectrum2D],
    Dict[str, List[WaveSpectrum1D]],
    Dict[str, List[WaveSpectrum2D]],
    Dict[str, List[List[WaveSpectrum2D]]],
    List[List[WaveSpectrum2D]],
    List[List[DataFrame]],
    Dict[int,List[DataFrame]],
    Dict[str,List[WaveBulkData]],
    Dict[str,DataFrame]
]


def _b64_encode_numpy(val:numpy.ndarray)->dict:
    if val.dtype == numpy.float64:
        val=val.astype(dtype=numpy.float32)

    if val.dtype == numpy.int64:
        val=val.astype(dtype=numpy.int32)

    dtype = val.dtype.str
    shape = val.shape
    data = base64.b64encode(val)

    return {"__class__":"ndarray", "data":{"dtype":dtype,"shape":shape,"data":data.decode('utf8')}}

def _b64_decode_numpy(data):
    decoded_bytes = base64.b64decode(bytes(data['data'],encoding='utf-8'))
    dtype = numpy.dtype(data['dtype'])
    array = numpy.frombuffer(decoded_bytes,dtype=dtype)
    return numpy.reshape(array,data['shape'])

class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, numpy.ndarray):
            return _b64_encode_numpy(obj)
        elif isinstance(obj, numpy.int32):
            return int(obj)
        elif isinstance(obj, numpy.float32):
            return float(obj)
        elif isinstance(obj, numpy.int64):
            return int(obj)
        elif isinstance(obj, numpy.float64):
            return float(obj)
        elif isinstance(obj, datetime):
            return {'__class__':'datetime', 'data':datetime.isoformat(obj)}
        elif isinstance(obj,DataFrame):
            return {'__class__':'DataFrame', 'data': obj.to_json(date_unit='s')}
        elif isinstance(obj, WaveSpectrum1D):
            return {'__class__': 'WaveSpectrum1D', 'data':obj._create_wave_spectrum_input()}
        elif isinstance(obj, WaveSpectrum2D):
            return {'__class__': 'WaveSpectrum2D', 'data':obj._create_wave_spectrum_input()}
        elif isinstance(obj, WaveBulkData):
            return {'__class__': 'WaveBulkData', 'data':obj.as_dict()}
        elif isinstance(obj, WindData):
            return {'__class__': 'WindData', 'data':obj.as_dict()}
        elif isinstance(obj, SSTData):
            return {'__class__': 'SSTData', 'data':obj.as_dict()}
        elif isinstance(obj, BarometricPressure):
            return {'__class__': 'BarometricPressure', 'data':obj.as_dict()}
        else:
            return json.JSONEncoder.default(self, obj)

def object_hook(dictionary:dict):
    if '__class__' in dictionary:
        if dictionary['__class__'] == 'DataFrame':
            #
            #print(dictionary['data'])
            df = read_json(dictionary['data'])
            if 'timestamp' in df:
                df['timestamp'] = df['timestamp'].apply(
                    lambda x: x.tz_localize('utc'))
            else:
                df.index = [ x.tz_localize('utc') for x in df.index]
            return df
        elif dictionary['__class__'] == 'WaveSpectrum1D':
            return spectrum1D(**dictionary['data'])
        elif dictionary['__class__'] == 'WaveSpectrum2D':
            return spectrum2D(**dictionary['data'])
        elif dictionary['__class__'] == 'ndarray':
            return _b64_decode_numpy(dictionary['data'])
        elif dictionary['__class__'] == 'datetime':
            return datetime.fromisoformat( dictionary['data'] )
        elif dictionary['__class__'] == 'WaveBulkData':
            return WaveBulkData(**dictionary['data'])
        elif dictionary['__class__'] == 'WaveBulkData':
            return WaveBulkData(**dictionary['data'])
        elif dictionary['__class__'] == 'WindData':
            return WindData(**dictionary['data'])
        elif dictionary['__class__'] == 'SSTData':
            return SSTData(**dictionary['data'])
        elif dictionary['__class__'] == 'BarometricPressure':
            return BarometricPressure(**dictionary['data'])
    else:
        # legacy - should delete soonish.
        if 'directions' in dictionary:
            if isinstance(dictionary['directions'],numpy.ndarray):
                return dictionary
            # if directions -> create a wavespectrum2d. The keys directly map
            # to the constructor arguments.
            return spectrum2D(**dictionary)
        elif 'frequency' in dictionary:
            if isinstance(dictionary['frequency'],numpy.ndarray):
                return dictionary
            # if not directions but has frequency -> create a wavespectrum1d.
            # The keys directly map to the constructor arguments.
            return spectrum1D(**dictionary)
        elif 'dataframe' in dictionary:
            df = read_json(dictionary['dataframe'])
            df['timestamp'] = df['timestamp'].apply(lambda x: x.tz_localize('utc'))
            return df #DataFrame.from_dict(data['dataframe'])
        return dictionary

def load(filename: str) -> _UNION:
    """
    Load spectral data as saved by "save_spectrum" from the given file and
    return a (nested) object. The precise format of the output depends on what
    was saved.

    :param filename: path to file to load.
    :return:
        - WaveSpectrum1D
          A single 1D spectrum

        - WaveSpectrum2D
          A single 2D spectrum

        - List[WaveSpectrum1D]
          List of 1D spectra

        - List[WaveSpectrum2D],
          List of 2D spectra.

        - Dict[str, List[WaveSpectrum1D]]
          Dictionary with Spotter/Observation name as keys, each containing a
          List of observed 1D spectra

        - Dict[str, List[WaveSpectrum2D]]
          Dictionary with Spotter/Observation name as keys, each containing a
          List of observed and reconstructed/computed 2D spectra

        - List[List[WaveSpectrum2D]]]
          List of partitioned data - each partition in itself is a list of
          2D wavespectra ordered in time.

        - Dict[str, List[List[WaveSpectrum2D]]]
          Dictionary with Spotter/Observation name as keys, each containing a list
          of partitioned data - each partition in itself is a list of
          2D wavespectra ordered in time.
    """

    with gzip.open(filename, 'rb') as file:
        data = file.read().decode('utf-8')
        return json.loads(data, object_hook=object_hook)


def save(_input: _UNION, filename: str, format='json', separate_spotters_into_files=False):
    """
    Save spectral data, possible in nested form as returned by the spectral
    partition/reconstruction/etc. functions. Data is saved in JSON form
    compressed with gzip.

    :param _input:
        - WaveSpectrum1D
          A single 1D spectrum

        - WaveSpectrum2D
          A single 2D spectrum

        - List[WaveSpectrum1D]
          List of 1D spectra

        - List[WaveSpectrum2D],
          List of 2D spectra.

        - Dict[str, List[WaveSpectrum1D]]
          Dictionary with Spotter/Observation name as keys, each containing a
          List of observed 1D spectra

        - Dict[str, List[WaveSpectrum2D]]
          Dictionary with Spotter/Observation name as keys, each containing a
          List of observed and reconstructed/computed 2D spectra

        - List[List[WaveSpectrum2D]]]
          List of partitioned data - each partition in itself is a list of
          2D wavespectra ordered in time.

        - Dict[str, List[List[WaveSpectrum2D]]]
          Dictionary with Spotter/Observation name as keys, each containing a list
          of partitioned data - each partition in itself is a list of
          2D wavespectra ordered in time.

    :param filename: path to save data.
    :return: None
    """

    def write(filename, _input, format):
        if format == 'json':
            with gzip.open(filename, 'wt') as file:
                json.dump(_input,file,cls=NumpyEncoder)
                #file.write(data.encode('utf-8'))
        else:
            raise Exception('Unknown output format')


    if separate_spotters_into_files:
        os.makedirs(filename, exist_ok=True)
        for key in _input:
            name = os.path.join(filename, key)
            write( name,_input[key],format )

    else:
        write(filename,_input,format)


import numpy
from typing import List
from .spectrum1D import WaveSpectrum1D, WaveSpectrum1DInput
from .spectrum2D import WaveSpectrum2D, WaveSpectrum2DInput
from datetime import timedelta

def spectrum1D_time_filter(spectra: List[WaveSpectrum1D],
                         window: numpy.ndarray = None,
                         maximum_time_delta: timedelta = None) -> List[
    WaveSpectrum1D]:
    number_of_spectra = len(spectra)

    if not window:
        window = numpy.array([0.25, 0.5, 0.25])

    if not maximum_time_delta:
        maximum_time_delta = timedelta(hours=2)

    window_length = len(window)
    window_midpoint = window_length // 2

    # shallow copy of the list
    spectra = spectra.copy()
    first = spectra[0]
    last = spectra[-1]

    for ii in range(0, window_midpoint):
        spectra.insert(0, first)
        spectra[0].timestamp = spectra[1].timestamp - maximum_time_delta

    for ii in range(window_midpoint + 1, window_length):
        spectra.append(last)
        spectra[-1].timestamp = spectra[-1].timestamp + maximum_time_delta

    output = []
    for index in range(0, number_of_spectra):
        istart = index
        iend = index + window_length

        timedeltas = numpy.array([
            next.timestamp - cur.timestamp for cur, next in
            zip(spectra[istart: iend - 1], spectra[istart + 1: iend])
        ])

        raw_spectrum = spectra[istart + window_midpoint].copy()
        if numpy.any(timedeltas > maximum_time_delta):
            output.append(raw_spectrum)
            continue

        e = sum(w * spectrum.e for w, spectrum in
                zip(window, spectra[istart: iend]))
        a1 = sum(w * spectrum.a1 * spectra[index].e for w, spectrum in
                 zip(window, spectra[istart: iend])) / e
        b1 = sum(w * spectrum.b1 * spectra[index].e for w, spectrum in
                 zip(window, spectra[istart: iend])) / e
        a2 = sum(w * spectrum.a2 * spectra[index].e for w, spectrum in
                 zip(window, spectra[istart: iend])) / e
        b2 = sum(w * spectrum.b2 * spectra[index].e for w, spectrum in
                 zip(window, spectra[istart: iend])) / e

        output.append(WaveSpectrum1D(WaveSpectrum1DInput(
            frequency=raw_spectrum.frequency,
            varianceDensity=e,
            timestamp=raw_spectrum.timestamp,
            latitude=raw_spectrum.latitude,
            longitude=raw_spectrum.longitude,
            a1=a1, b1=b1,
            a2=a2, b2=b2))
        )
    return spectra


def spectrum2D_time_filter(spectra: List[WaveSpectrum2D],
                         window: numpy.ndarray = None,
                         maximum_time_delta: timedelta = None) -> List[
    WaveSpectrum2D]:
    number_of_spectra = len(spectra)

    if not window:
        window = numpy.array([0.25, 0.5, 0.25])

    if not maximum_time_delta:
        maximum_time_delta = timedelta(hours=2)

    window_length = len(window)
    window_midpoint = window_length // 2

    # shallow copy of the list
    spectra = spectra.copy()
    first = spectra[0]
    last = spectra[-1]

    for ii in range(0, window_midpoint):
        spectra.insert(0, first)
        spectra[0].timestamp = spectra[1].timestamp - maximum_time_delta

    for ii in range(window_midpoint + 1, window_length):
        spectra.append(last)
        spectra[-1].timestamp = spectra[-1].timestamp + maximum_time_delta

    output = []
    for index in range(0, number_of_spectra):
        istart = index
        iend = index + window_length

        timedeltas = numpy.array([
            next.timestamp - cur.timestamp for cur, next in
            zip(spectra[istart: iend - 1], spectra[istart + 1: iend])
        ])

        raw_spectrum = spectra[istart + window_midpoint].copy()
        if numpy.any(timedeltas > maximum_time_delta):
            output.append(raw_spectrum)
            continue

        variance_density = sum(w * spectrum.variance_density for w, spectrum in
                zip(window, spectra[istart: iend]))

        output.append(
            WaveSpectrum2D(
                WaveSpectrum2DInput(
                    frequency=raw_spectrum.frequency,
                    varianceDensity=variance_density,
                    timestamp=raw_spectrum.timestamp,
                    latitude=raw_spectrum.latitude,
                    longitude=raw_spectrum.longitude,
                    directions=raw_spectrum.direction
                )
            )
        )
    return spectra


from dataclasses import dataclass, field
from datetime import datetime, timedelta
from .timebase import ModelTimeConfiguration, timebase_forecast, timebase_lead, \
    timebase_evaluation
import os
from typing import List,Tuple
import json

with open( os.path.expanduser('~/model_configuration.json'),'rb') as file:
    MODELS = json.load(file)

def available_models():
    return list(MODELS.keys())

@dataclass()
class Model():
    key:str
    name: str
    label: str
    bucket: str
    resolution: str
    region: str
    key_template: str
    filetype:str
    model_time_configuration: ModelTimeConfiguration
    _model_time_configuration: ModelTimeConfiguration =  field(init=False, repr=False)

    @property
    def model_time_configuration(self) -> ModelTimeConfiguration:
        return self._model_time_configuration

    @model_time_configuration.setter
    def model_time_configuration(self, values):
        self._model_time_configuration = ModelTimeConfiguration(**values)


def get_model(name):
    return Model(key=name,**MODELS[name])


def _replace(string: str, properties):
    for key, value in properties.items():
        if key in string:
            string = string.replace("{" + key + "}", value)
    return string

#ecmwf/{prefix}{date_string}{hour_string}00{valid_time_date}{valid_time_hour}011
def generate_aws_key(variable, init_time: datetime, forecast_hour: timedelta,
                     model: Model):
    #
    if forecast_hour == timedelta(hours=0):
        ecmwf_lead_zero_indicator = '011'
    else:
        ecmwf_lead_zero_indicator = '001'

    if (init_time.hour == 0) or (init_time.hour == 12):
        ecmwf_subcycle_string = 'A2P'
    else:
        ecmwf_subcycle_string = 'A2Q'
    valid_time = init_time+forecast_hour

    properties = {
        "name": model.name,
        "label": model.label,
        "bucket": model.bucket,
        "resolution": model.resolution,
        "region": model.region,
        "date": init_time.strftime("%Y%m%d"),
        "cyclehour": init_time.strftime("%H"),
        "forecasthour": "{hours:03d}".format(
            hours=int(forecast_hour.total_seconds() // 3600)),
        "variable": variable,
        "key":model.key,
        "init_time_month": init_time.strftime("%m"),
        "init_time_day": init_time.strftime("%d"),
        "valid_time_month": valid_time.strftime("%m"),
        "valid_time_day": valid_time.strftime("%d"),
        "valid_time_hour": valid_time.strftime("%H"),
        "ecmwf_subcycle_string":ecmwf_subcycle_string,
        "ecmwf_lead_zero_indicator":ecmwf_lead_zero_indicator
    }
    key = _replace(model.key_template, properties)

    return f"{model.bucket}/{key}"


def generate_forecast_keys_and_valid_times(variable, init_time: datetime, duration: timedelta,
                           model: Model) -> Tuple[List[str],List[datetime]] :

    time_vector = timebase_forecast(init_time, duration,
                                    time_configuration=model.model_time_configuration)

    aws_keys = []
    valid_time = []
    for time in time_vector:
        aws_keys.append(
            generate_aws_key(variable, init_time=time[0],
                             forecast_hour=time[1], model=model)
        )
        valid_time.append(time[0]+time[1])
    return aws_keys, valid_time


def generate_lead_keys_and_valid_times(variable, start_time: datetime, end_time: datetime,
                            lead_time: timedelta, model: Model, exact=False) -> Tuple[List[str],List[datetime]]:

    time_vector = timebase_lead(start_time, end_time, lead_time,
                                time_configuration=model.model_time_configuration,
                                exact=exact)

    aws_keys = []
    valid_time = []
    for time in time_vector:
        aws_keys.append(
            generate_aws_key(variable, init_time=time[0],
                             forecast_hour=time[1], model=model)
        )
        valid_time.append(time[0] + time[1])
    return aws_keys, valid_time



from datetime import datetime, timezone, timedelta
from dataclasses import dataclass, field
from typing import List, Union, Tuple
import numpy

INTERVALTYPE = Union[List[tuple[int, timedelta]], Tuple[tuple[int, timedelta]]]


def to_timedelta(value):
    if isinstance(value, timedelta):
        return value
    else:
        return timedelta(hours=int(value))


def find_closet_init_time(evaluation_time: datetime,
                          cycle_time_hours: timedelta,
                          cycle_offset_hours: timedelta,
                          lead_time: timedelta = timedelta(hours=0)):
    evaluation_time_seconds = evaluation_time.timestamp()
    cycle_time_seconds = cycle_time_hours.total_seconds()
    cycle_offset_seconds = cycle_offset_hours.total_seconds()
    lead_time_seconds = lead_time.total_seconds()

    number_of_cycles = (evaluation_time_seconds - cycle_offset_seconds- lead_time_seconds) // \
                       cycle_time_seconds



    init_time = datetime.fromtimestamp(
        number_of_cycles * cycle_time_seconds
            + cycle_offset_seconds,
        tz=timezone.utc
    )
    delta = evaluation_time_seconds - lead_time_seconds - init_time.timestamp()

    if delta >= cycle_time_seconds/2:
        new_lead_time = evaluation_time - init_time - cycle_time_hours
        if new_lead_time.total_seconds() >= 0:
            init_time = init_time + cycle_time_hours

    #print( (lead_time_seconds - cycle_time_seconds * (lead_time_seconds//cycle_time_seconds))/3600 //2)
    #
    # actual_lead = evaluation_time - init_time
    # lead_delta  = int(lead_time.total_seconds()-actual_lead.total_seconds()/3600)
    #
    # number_of_cycles = numpy.round(lead_delta/cycle_time_seconds,0)
    # init_time = init_time-number_of_cycles*cycle_time_hours
    return init_time


class ModelTimeConfiguration():
    def __init__(self,
                 cycle_time_hours: timedelta = timedelta(hours=6),
                 cycle_offset_hours: timedelta = timedelta(hours=0),
                 output_interval: INTERVALTYPE = (
                         (240, timedelta(hours=1)),)):

        self.cycle_time_hours = cycle_time_hours
        self.cycle_offset_hours = cycle_offset_hours
        self.output_interval = output_interval

    @property
    def cycle_time_hours(self):
        return self._cycle_time_hours

    @cycle_time_hours.setter
    def cycle_time_hours(self, value):
        self._cycle_time_hours = to_timedelta(value)

    @property
    def cycle_offset_hours(self):
        return self._cycle_offset_hours

    @cycle_offset_hours.setter
    def cycle_offset_hours(self, value):
        self._cycle_offset_hours = to_timedelta(value)

    @property
    def output_interval(self) -> INTERVALTYPE:
        return self._output_interval

    @output_interval.setter
    def output_interval(self, values):
        self._output_interval = []
        for value in values:
            if isinstance(value[1], timedelta):
                self._output_interval.append(value)
            else:
                self._output_interval.append(
                    (value[0], timedelta(hours=value[1])))


def timebase_forecast(init_time: datetime, duration: timedelta,
                      time_configuration: ModelTimeConfiguration):
    timebase = [(init_time, timedelta(hours=0))]
    for intervals in time_configuration.output_interval:
        #
        base_hour = timebase[-1][1]
        for index in range(intervals[0]):
            timebase.append(
                (init_time, base_hour + intervals[1] * (index + 1)))

            if intervals[1] * index >= duration:
                return timebase

    return timebase


def timebase_lead(start_time: datetime, end_time: datetime,lead_time: timedelta,
                  time_configuration: ModelTimeConfiguration, exact=True):

    n = int((end_time-start_time)/time_configuration.output_interval[0][1])
    times = []
    for index in range(0, n):
        #
        valid_time = start_time + index * time_configuration.output_interval[0][1]
        init_time = find_closet_init_time(valid_time,
                                      cycle_time_hours=time_configuration.cycle_time_hours,
                                      cycle_offset_hours=time_configuration.cycle_offset_hours,
                                      lead_time=lead_time)
        forecast_hours = valid_time-init_time
        if exact:
            if not (forecast_hours==lead_time):
                continue

        times.append( (init_time,forecast_hours) )

    return times


def timebase_evaluation(start_time: datetime, end_time: datetime,
                        time_configuration: ModelTimeConfiguration):
    pass



"""
Contents: Routines to get spectral data from the spotter api

Copyright (C) 2022
Sofar Ocean Technologies

Authors: Pieter Bart Smit
======================

Routines to get spectral data from the spotter api

Functions:

- `get_spectrum`, function to call externally do download
                                         spectral data. Parallel if multiple Spotters
                                         are given.
- `_get_next_100_spectra`, Internal helper function. Grabs the next 100 spectra
                          from the given start data. Needed because the Spotter_API
                          will return a maximum of 100 spectra per call.
- '_download_spectra', function that abstracts away the limitation of 100 spectra
                       per call. Handles updating the start date etc.

How To Use This Module
======================
(See the individual functions for details.)

1. Import it: ``import roguewave.externaldata.spotterapi``
2. call ``get_spectrum`` to get the desired data.

TOC
======================

1) Imports: all external import data
2) Constants: all constants used in the module
3) Interfaces: overloaded interfaces for the main function (handles typing, no
               actual logic)
4) Main function: implementation of the main function
    - get_spectrum
5) Private Functions: module private functions that are used by the main function:
    - _get_next_100_spectra
    - _download_spectra
"""

# 1) Imports
# ======================
from datetime import datetime, timedelta, timezone
from .exceptions import \
    ExceptionNoDataForVariable
from multiprocessing.pool import ThreadPool
from pysofar.spotter import Spotter, SofarApi
from roguewave import logger
from roguewave.tools import datetime_to_iso_time_string, to_datetime
from roguewave.wavespectra.spectrum1D import WaveSpectrum1D, \
    WaveSpectrum1DInput
from roguewave.metoceandata import WaveBulkData, as_dataframe, WindData, \
    SSTData, MetoceanData, BarometricPressure
from typing import Dict, List, Union, overload, TypedDict, Tuple
from tqdm import tqdm
from pandas import DataFrame
import numpy

API = None

# 2) Constants
# ======================

# Maximum number of spectra to retrieve from the Spotter API per API call. Note
# that 2- os a hard limit of the API. If set higher than 100 it will just return
# 100 (and the implementation will fail)
MAX_LOCAL_LIMIT = 100
MAX_LOCAL_LIMIT_BULK = 500

# Maximum number of workers in the Threadpool. Should be set to something reasonable
# to not overload wavefleet
MAXIMUM_NUMBER_OF_WORKERS = 40
NUMBER_OF_RETRIES = 2


# 2.5 API Return JSON format:
class ApiWaveData(TypedDict):
    significantWaveHeight: float
    peakPeriod: float
    meanPeriod: float
    peakDirection: float
    peakDirectionalSpread: float
    meanDirection: float
    meanDirectionalSpread: float
    timestamp: str
    latitude: float
    longitude: float


class ApiWindData(TypedDict):
    speed: float
    direction: float
    timestamp: str
    latitude: float
    longitude: float


class ApiSSTData(TypedDict):
    degrees: float
    timestamp: str
    latitude: float
    longitude: float


class VariablesToInclude(TypedDict):
    frequencyData: bool
    waves: bool
    wind: bool
    surfaceTemp: bool
    barometerData: bool


# 3) Interfaces
# ======================
# these overloads only exist so we can make use of typing and autocomplete in
# e.g. Pycharm. Specifically, it handles the case that the return type depends
# on the input.
@overload
def get_spectrum(
        spotter_ids: List[str],
        start_date: Union[datetime, int, float, str] = None,
        end_date: Union[datetime, int, float, str] = None,
        session: SofarApi = None,
        parallel_download=True
) -> Dict[str, List[WaveSpectrum1D]]: ...


@overload
def get_spectrum(
        spotter_ids: str,
        start_date: Union[datetime, int, float, str] = None,
        end_date: Union[datetime, int, float, str] = None,
        session: SofarApi = None,
        parallel_download=True
) -> List[WaveSpectrum1D]: ...


@overload
def get_bulk_wave_data(
        spotter_ids: List[str],
        start_date: Union[datetime, int, float, str] = None,
        end_date: Union[datetime, int, float, str] = None,
        session: SofarApi = None,
        parallel_download=True
) -> Dict[str, List[WaveBulkData]]: ...


@overload
def get_bulk_wave_data(
        spotter_ids: str,
        start_date: Union[datetime, int, float, str] = None,
        end_date: Union[datetime, int, float, str] = None,
        session: SofarApi = None,
        parallel_download=True
) -> List[WaveBulkData]: ...


# 4) Main Function
# ======================
# Implements the interfaces above.
def get_sofar_api():
    if API is None:
        return SofarApi()
    else:
        return API


def get_spotter_ids(sofar_api: SofarApi = None) -> List[str]:
    if sofar_api is None:
        sofar_api = get_sofar_api()
    return sofar_api.device_ids


def get_spectrum(
        spotter_ids: Union[str, List[str]],
        start_date: Union[datetime, int, float, str] = None,
        end_date: Union[datetime, int, float, str] = None,
        session: SofarApi = None,
        parallel_download=True
) -> Union[Dict[str, List[WaveSpectrum1D]], List[WaveSpectrum1D]]:
    """
    Grabs the requested spectra for the spotter(s) in the given interval

    :param spotter_ids:
        Can be either 1) a List of spotter_ids or 2) a single Spotter_id.
            If a List of Spotters, i.e.: List[str]:
                the return type is a Dictionary, with the spotter_id as key and
                the dictionary entry a list of spectra for that Spotter in the
                requested time frame -> Dict[str,List[WaveSpectrum1D]]

            If a single spotter, i.e.: str:
                the return type is a Dictionary, with the spotter_id as key and
                the dictionary entry a list of spectra for that Spotter in the
                requested time frame. List[WaveSpectrum1D]

    :param start_date: ISO 8601 formatted date string, epoch or datetime.
                       If not included defaults to beginning of spotters history
    :param end_date: ISO 8601 formatted date string, epoch or datetime.
                     If not included defaults to end of spotter history

    :return: Data as a FrequencyDataList Object
    """
    data = get_data(
        spotter_ids,
        start_date,
        end_date,
        include_frequency_data=True,
        include_directional_moments=True,
        include_waves=False,
        include_wind=False,
        include_surface_temp_data=False,
        session=session,
        parallel_download=parallel_download,
    )
    out = {}
    for key in data:
        out[key] = data[key]['frequencyData']
    return out


def get_bulk_wave_data(
        spotter_ids: Union[str, List[str]],
        start_date: Union[datetime, int, float, str] = None,
        end_date: Union[datetime, int, float, str] = None,
        session: SofarApi = None,
        parallel_download=True
) -> Union[Dict[str, List[WaveBulkData]], List[WaveBulkData]]:
    """
    Grabs the requested spectra for the spotter(s) in the given interval

    :param spotter_ids:
        Can be either 1) a List of spotter_ids or 2) a single Spotter_id.
            If a List of Spotters, i.e.: List[str]:
                the return type is a Dictionary, with the spotter_id as key and
                the dictionary entry a list of spectra for that Spotter in the
                requested time frame -> Dict[str,List[WaveSpectrum1D]]

            If a single spotter, i.e.: str:
                the return type is a Dictionary, with the spotter_id as key and
                the dictionary entry a list of spectra for that Spotter in the
                requested time frame. List[WaveSpectrum1D]

    :param start_date: ISO 8601 formatted date string, epoch or datetime.
                       If not included defaults to beginning of spotters history
    :param end_date: ISO 8601 formatted date string, epoch or datetime.
                     If not included defaults to end of spotter history

    :return: Data as a FrequencyDataList Object
    """
    data = get_data(
        spotter_ids,
        start_date,
        end_date,
        include_frequency_data=False,
        include_directional_moments=False,
        include_waves=True,
        include_wind=False,
        include_surface_temp_data=False,
        session=session,
        parallel_download=parallel_download
    )
    out = {}
    for key in data:
        out[key] = data[key]['waves']
    return out


def get_data(
        spotter_ids: Union[str, List[str]],
        start_date: Union[datetime, int, float, str] = None,
        end_date: Union[datetime, int, float, str] = None,
        include_frequency_data=False,
        include_directional_moments=False,
        include_waves=True,
        include_wind=False,
        include_barometer_data=False,
        include_surface_temp_data=False,
        session: SofarApi = None,
        parallel_download=True,
        bulk_data_as_dataframe=True
) -> Dict[str, Dict[
    str, Union[list[WaveSpectrum1D], list[WaveBulkData], DataFrame]]]:
    if spotter_ids is None:
        spotter_ids = get_spotter_ids()

    variables_to_include = VariablesToInclude(
        waves=include_waves,
        wind=include_wind,
        surfaceTemp=include_surface_temp_data,
        frequencyData=include_frequency_data,
        barometerData=include_barometer_data,
    )

    if not isinstance(spotter_ids, list):
        spotter_ids = [spotter_ids]
        return_list = False
    else:
        return_list = True

    if session is None:
        session = SofarApi()

    def worker(spotter_id):
        data = _download_data(spotter_id, session, variables_to_include,
                              start_date, end_date, bulk_data_as_dataframe)
        return data

    if parallel_download:
        with ThreadPool(processes=MAXIMUM_NUMBER_OF_WORKERS) as pool:
            out = list(
                tqdm(pool.imap(worker, spotter_ids),
                     total=len(spotter_ids)))
    else:
        out = list(
            tqdm(map(worker, spotter_ids), total=len(spotter_ids)))

    data = {}
    for spotter_id, spotter_data in zip(spotter_ids, out):
        #
        # Did we get any data for this spotter
        if spotter_data is not None:
            data[spotter_id] = spotter_data

    if not return_list:
        return data[spotter_ids[0]]
    else:
        return data


# 5) Private Functions
# ======================
def _download_data(
        spotter_id: str,
        session: SofarApi,
        variables_to_include: VariablesToInclude,
        start_date: Union[datetime, str, int, float] = None,
        end_date: Union[datetime, str, int, float] = None,
        bulk_data_as_dataframe: bool = True,
        limit: int = None) -> Dict[
    str, Union[List[WaveSpectrum1D], List[WaveBulkData]]]:
    """
    Function that downloads data from the API for the requested Spotter
    It abstracts away the limitation that the API can only return a maximum
    of 100 Spectra or 500 bulk data points for a single Spotter per call.

    :param spotter_id: ID for the Spotter we want to download
    :param session: Active session of the SofarAPI
    :param start_date: ISO 8601 formatted date string, epoch or datetime.
                       If not included defaults to beginning of spotters history
    :param end_date: ISO 8601 formatted date string, epoch or datetime.
                     If not included defaults to end of spotter history
    :param limit: Maximum number of Spectra to download per call. Not exposed
                  externally right now as there is not really a reason to change
                  it. (calling function does not set the limit)

    :return: List of available data in the requested timeframe.
    """

    # Create a Spotter object to Query
    spotter = Spotter(spotter_id, spotter_id, session=session)

    # Set the initial start date. This will get advanced for every 100 spectra
    # we download
    _start_date = start_date

    if variables_to_include['frequencyData']:
        max_local_limit = MAX_LOCAL_LIMIT
    else:
        max_local_limit = MAX_LOCAL_LIMIT_BULK

    data = {}
    while True:
        # We can only download a maximum of 100 spectra at a time; so we need
        # to loop our request. We do not know how many spotters there are
        # in the given timeframe.
        #
        # Assumptions:
        #   - spotter api returns a maximum of N items per requests (N=500 if
        #     no spectral data, N=100 if so)
        #   - requests returned start from the requested start data and
        #     with the last entry either being the last entry that fits
        #     in the requested window, or merely the last sample that fits
        #     in the 100 items.
        #   - requests returned are in order

        number_of_items_returned = 0
        if limit is not None:
            # If we hit the limit (if given) of spotters requested, break
            if len(data) >= limit:
                break

            # otherwise, update our next request so we don't overrun
            # the limit
            local_limit = min(limit - len(data), max_local_limit)
        else:
            # if no limit is given, just ask for the maximum allowed.
            local_limit = max_local_limit

        try:
            # Try to get the next batch of data
            next, number_of_items_returned, max_timestamp = \
                _get_next_page(spotter, variables_to_include,
                               _start_date, end_date, local_limit)

        except ExceptionNoDataForVariable as e:
            # Could not download data, add nothing, raise warning
            raise e

        # Add the data to the list

        for key in next:
            if key in data:
                data[key] += next[key]
            else:
                data[key] = next[key]

        # If we did not receive all data we requested...
        if number_of_items_returned < local_limit:
            # , we are done...
            break
        else:
            # ... else we update the startdate to be the timestamp of the last
            # known entry we received plus a second, and use this as the new
            # start.
            _start_date = max_timestamp + timedelta(seconds=1)

    # Postprocessing
    if len(data) < 1:
        data = None
    else:
        # Convert bulk data to a dataframe if desired
        for key in data:
            if key == 'frequencyData':
                # We cannot convert the list of wavespectra to a dataframe.
                continue
            data[key] = as_dataframe(data[key])

    return data


def _get_next_page(
        spotter: Spotter,
        variables_to_include: VariablesToInclude,
        start_date: Union[datetime, str, int, float] = None,
        end_date: Union[datetime, str, int, float] = None,
        limit: int = MAX_LOCAL_LIMIT,
) -> Tuple[
    Dict[str, Union[List[WaveSpectrum1D], List[WaveBulkData]]], int, datetime]:
    """
    Function that downloads the page of Data from the Spotter API that lie
    within the given interval, starting from the record closest to the startdate.

    idiosyncrasies to handle:
    - Wavefleet sometimes returns multiple instances of the same record (same
      timestamp). These are filtered through a all to _unique. This should
      be fixed in the future.
    - Some entries are broken (None for entries).
    - Not all Spotters will have (all) data for the given timerange.
    - Wavefleet sometimes times out on requests.

    :param spotter: Spotter object from pysofar

    :param start_date: ISO 8601 formatted date string, epoch or datetime.
                       If not included defaults to beginning of spotters history

    :param end_date: ISO 8601 formatted date string, epoch or datetime.
                     If not included defaults to end of spotter history

    :param limit: Maximum number of Spectra to download per call. Not exposed
                  externally right now as there is not really a reason to change
                  it. (calling function does not set the limit)

    :return: Data
    """

    # Retry mechanism
    start_date = to_datetime(start_date)
    end_date = to_datetime(end_date)
    json_data = None
    for retry in range(0, NUMBER_OF_RETRIES + 1):
        try:
            json_data = spotter.grab_data(
                limit=limit,
                start_date=datetime_to_iso_time_string(start_date),
                end_date=datetime_to_iso_time_string(end_date),
                include_frequency_data=variables_to_include[
                    'frequencyData'],
                include_directional_moments=variables_to_include[
                    'frequencyData'],
                include_barometer_data=variables_to_include['barometerData'],
                include_waves=variables_to_include['waves'],
                include_wind=variables_to_include['wind'],
                include_surface_temp_data=variables_to_include[
                    'surfaceTemp'],
            )
            break
        except Exception as e:
            if retry < NUMBER_OF_RETRIES:
                warning = f'Error downloading data for {spotter.id}, attempting retry {retry + 1}'
                logger.warning(warning)
            else:
                raise Exception

    out = {}
    max_num_items = 0
    max_timestamp = datetime(1970, 1, 1, tzinfo=timezone.utc)
    #
    # Loop over all possible variables
    for var_name, include_variable in variables_to_include.items():
        #
        # Did we want to include this variable?
        if include_variable:
            #
            # If so- was it returned? If not- raise error
            if var_name not in json_data:
                raise ExceptionNoDataForVariable(var_name)

            # If so- were any elements returned for this period? If not continue
            # We could error here, but there may be gaps in data for certain
            # sensors, so we try to continue
            if not json_data[var_name]:
                continue

            elements_returned = len(json_data[var_name])

            # Filter for Nones.
            json_data[var_name] = _none_filter(json_data[var_name])
            if not json_data[var_name]:
                # If after filtering for bad data nothing remains, continue.
                continue

            # How many items were returned (counting doubles and none values,
            # this is used for the pagination logic which only knows if it is
            # done if less then the requested data was returned).
            max_num_items = max(max_num_items, elements_returned)

            # Filter for doubles.
            json_data[var_name] = _unique_filter(json_data[var_name])

            # Add to output
            out[var_name] = \
                [_get_class(var_name, data) for data in json_data[var_name]]

            if out[var_name][-1].timestamp > max_timestamp:
                max_timestamp = out[var_name][-1].timestamp

    return out, max_num_items, max_timestamp


def search_circle(
        start_date: Union[datetime, str],
        end_date: Union[datetime, str],
        center_lat_lon: Tuple,
        radius: float,
        session: SofarApi=None,
        variables_to_include: VariablesToInclude = None,
        page_size=500,
        bulk_data_as_dataframe: bool = True
):
    geometry = {'type': 'circle', 'points': center_lat_lon, 'radius': radius}

    return search(start_date, end_date, geometry, session,
                  variables_to_include, page_size, bulk_data_as_dataframe)


def search_rectangle(
        start_date: Union[datetime, str],
        end_date: Union[datetime, str],
        north_west_lat_lon,
        south_east_lat_lon,
        session: SofarApi=None,
        variables_to_include: VariablesToInclude = None,
        page_size=500,
        bulk_data_as_dataframe: bool = True
):
    geometry = {'type': 'envelope',
                'points': [north_west_lat_lon, south_east_lat_lon],
                'radius': None}

    return search(start_date, end_date, geometry, session,
                  variables_to_include, page_size, bulk_data_as_dataframe)


def search(start_date: Union[datetime, str],
           end_date: Union[datetime, str],
           geometry: dict,
           session: SofarApi=None,
           variables_to_include: VariablesToInclude = None,
           page_size=500,
           bulk_data_as_dataframe: bool = True
           ):
    if session is None:
        session = SofarApi()

    if variables_to_include is None:
        variables_to_include = VariablesToInclude(
            frequencyData=True, waves=True, wind=True, surfaceTemp=True,
            barometerData=True
        )

    start_date_str = datetime_to_iso_time_string(start_date)
    end_date_str = datetime_to_iso_time_string(end_date)

    shape = geometry['type']
    shape_params = geometry['points']
    radius = geometry['radius']

    generator = session.search(
        shape=shape,
        shape_params=shape_params,
        start_date=start_date_str,
        end_date=end_date_str,
        radius=radius,
        page_size=page_size,
        return_generator=True
    )

    spotters = {}
    # loop over all spotters returned
    for spotter in generator:
        spotter_id = spotter['spotterId']
        # loop over keys we can parse
        for key in variables_to_include:
            if key in spotter:
                item = spotter[key]
                if not item:
                    # no data
                    continue

                item['latitude'] = spotter['latitude']
                item['longitude'] = spotter['longitude']
                item['timestamp'] = spotter['timestamp']
                data = _get_class(key, item)
                if spotter_id not in spotters:
                    spotters[spotter_id] = {}

                if key not in spotters[spotter_id]:
                    spotters[spotter_id][key] = []

                spotters[spotter_id][key].append(data)


    for spotter_id in spotters:
        for key in spotters[spotter_id]:
            if bulk_data_as_dataframe and (not key == 'frequencyData'):
                spotters[spotter_id][key]= as_dataframe(spotters[spotter_id][key])
    return spotters


# 6) Helper Functions
# ======================
def _unique_filter(data):
    """
    Filter for dual time entries that occur due to bugs in wavefleet (same
    record returned twice)
    :param data:
    :return:
    """
    timestamps = numpy.array(
        [to_datetime(x['timestamp']).timestamp() for x in data])
    _, unique_indices = numpy.unique(timestamps, return_index=True)

    data = [data[index] for index in unique_indices]

    return data


def _none_filter(data: List[Union[WaveSpectrum1D, WaveBulkData]]):
    F = lambda x: (x['latitude'] is not None) and (
            x['longitude'] is not None) and \
                  (x['timestamp'] is not None)

    return list(filter(F, data))


def _get_class(key, data) -> Union[MetoceanData, WaveSpectrum1D]:
    if key == 'waves':
        return WaveBulkData(
            timestamp=data['timestamp'],
            latitude=data['latitude'],
            longitude=data['longitude'],
            significant_waveheight=data['significantWaveHeight'],
            peak_period=data['peakPeriod'],
            mean_period=data['meanPeriod'],
            peak_direction=data['peakDirection'],
            peak_directional_spread=data['peakDirectionalSpread'],
            mean_direction=data['meanDirection'],
            mean_directional_spread=data['meanDirectionalSpread'],
            peak_frequency=1.0 / data['peakPeriod']
        )
    elif key == 'frequencyData':
        return WaveSpectrum1D(WaveSpectrum1DInput(**data))
    elif key == 'wind':
        return WindData(**data)
    elif key == 'surfaceTemp':
        return SSTData(**data)
    elif key == 'barometerData':
        return BarometricPressure(**data)
    else:
        raise Exception('Unknown variable')
